/*************************************************************************************************
   Demo / rzrDemo.cpp

   - Demo system

*************************************************************************************************/
#include "stdafx.h"
#include "../directx/Include/d3dx9.h"

// Incbin
#include "rzrLogo.h"
#include "rzrLogoa.h"
#include "rgbmix.h"

#define D3DSafeRelease(a) if (a) {a->Release(); a = rzNULL;}

#define KeyFrame0 1.5f
#define KeyFrame1 6.5f
#define KeyFrame2 8.f

#ifdef FMUSIC_ALL_ACTIVE
#include <stdio.h>
#else
#include "module.h"
#endif

struct MEMFILE
{
    int length;
    int pos;
    void *data;
};

unsigned int memopen(char *name)
{
#ifdef FMUSIC_ALL_ACTIVE
    return (unsigned int)fopen(name, "rb");
#else
    MEMFILE *pMemfile;

    pMemfile = new MEMFILE;
    pMemfile->data = module;
    pMemfile->length = sizeof(module);
    pMemfile->pos = 0;

    return (unsigned int)pMemfile;
#endif
}

void memclose(unsigned int handle)
{
#ifdef FMUSIC_ALL_ACTIVE
    fclose((FILE*)handle);
#else
    MEMFILE* pMemfile = (MEMFILE*)handle;
    delete pMemfile;
#endif
}

int memread(void *buffer, int size, unsigned int handle)
{
#ifdef FMUSIC_ALL_ACTIVE
    return fread(buffer, size, 1, (FILE*)handle);
#else
    MEMFILE *pMemfile = (MEMFILE*)handle;

    if (pMemfile->pos + size>=pMemfile->length)
        size = pMemfile->length - pMemfile->pos;

    memcpy(buffer, (char*)pMemfile->data + pMemfile->pos, size);
    pMemfile->pos += size;

    return size;
#endif
}

void memseek(unsigned int handle, int pos, signed char mode)
{
#ifdef FMUSIC_ALL_ACTIVE
    fseek((FILE*)handle, pos, mode);
#else
    MEMFILE *pMemfile = (MEMFILE *)handle;

    if (mode==0) //SEEK_SET
        pMemfile->pos = pos;
    else if (mode==1) //SEEK_CUR
        pMemfile->pos += pos;
    else if (mode==2) //SEEK_END
        pMemfile->pos = pMemfile->length + pos;

    if (pMemfile->pos>pMemfile->length)
        pMemfile->pos = pMemfile->length;
#endif
}

int memtell(unsigned int handle)
{
#ifdef FMUSIC_ALL_ACTIVE
    return ftell((FILE*)handle);
#else
    MEMFILE *pMemfile = (MEMFILE *)handle;

    return pMemfile->pos;
#endif
}

/*************************************************************************************************
 *
 * Constructor
 */
rzrDemo::rzrDemo():
    m_bFirstFrame(true), m_fTime(0.f),
    m_pD3D(rzNULL), m_pD3DDevice(rzNULL), m_pD3DScreen(rzNULL), m_pD3DVD(rzNULL), m_pD3DPS(rzNULL), m_pD3DTextureLogo(rzNULL),
    m_hBitmap(rzNULL),
    m_pMod(rzNULL)
{
    QueryPerformanceFrequency(&m_qPerfFrequency);
}

/*************************************************************************************************
 *
 * Destructor
 */
rzrDemo::~rzrDemo()
{
    if (m_pMod)
        FMUSIC_FreeSong(m_pMod);
    if (m_hBitmap)
        DeleteObject(m_hBitmap);

    D3DSafeRelease(m_pD3DTextureLogo);
    D3DSafeRelease(m_pD3DPS);
    D3DSafeRelease(m_pD3DVD);
    D3DSafeRelease(m_pD3DScreen);
    D3DSafeRelease(m_pD3DDevice);
    D3DSafeRelease(m_pD3D);
}

/*************************************************************************************************
 *
 * Init the demo
 */
rzErrCode rzrDemo::Init(HWND hWnd)
{
    m_hWnd = hWnd;
    // Force the window to be transparent and remove it from the taskbar
    SetWindowLong(m_hWnd, GWL_EXSTYLE, WS_EX_LAYERED|WS_EX_TOOLWINDOW);

    m_pD3D = Direct3DCreate9(D3D_SDK_VERSION);
    if (!m_pD3D)
        return rzErrFail;

    // Set up the structure used to create the D3DDevice
    D3DPRESENT_PARAMETERS d3dpp;
    ZeroMemory(&d3dpp, sizeof(d3dpp));

    d3dpp.Windowed = TRUE;
    d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;

    // Create the D3DDevice
    if (FAILED(m_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_SOFTWARE_VERTEXPROCESSING, &d3dpp, &m_pD3DDevice)))
    {
        return rzErrFail;
    }
    m_dWidth = d3dpp.BackBufferWidth;
    m_dHeight = d3dpp.BackBufferHeight;

    // Vertex declaration for the draw
    D3DVERTEXELEMENT9 pVertexElem[] =
    {
        {0, 0, D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITIONT, 0},
        {0, 16, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0},
        {0, 24, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 1},
        {0, 32, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 2},
        {0, 40, D3DDECLTYPE_FLOAT1, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 3},
        D3DDECL_END()
    };
    m_pD3DDevice->CreateVertexDeclaration(pVertexElem, &m_pD3DVD);

    m_pD3DDevice->CreatePixelShader((DWORD*)g_ps20_PS, &m_pD3DPS);

    // Load the logo
    D3DXIMAGE_INFO d3dxImageInfo;
    if (FAILED(D3DXCreateTextureFromFileInMemoryEx(m_pD3DDevice, rzrLogo, sizeof(rzrLogo), D3DX_DEFAULT, D3DX_DEFAULT, 1,
        0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, D3DX_DEFAULT, D3DX_DEFAULT, 0, &d3dxImageInfo, rzNULL, &m_pD3DTextureLogo)))
    {
        return rzErrFail;
    }
    m_dLogoWidth = d3dxImageInfo.Width;
    m_dLogoHeight = d3dxImageInfo.Height;

    D3DSURFACE_DESC d3dDesc;
    m_pD3DTextureLogo->GetLevelDesc(0, &d3dDesc);
    m_dTextureLogoWidth = d3dDesc.Width;
    m_dTextureLogoHeight = d3dDesc.Height;

    IDirect3DTexture9* pD3DTexture = rzNULL;
    if (FAILED(D3DXCreateTextureFromFileInMemoryEx(m_pD3DDevice, rzrLogoa, sizeof(rzrLogoa), D3DX_DEFAULT, D3DX_DEFAULT, 1,
        0, D3DFMT_A8R8G8B8, D3DPOOL_SYSTEMMEM, D3DX_DEFAULT, D3DX_DEFAULT, 0, rzNULL, rzNULL, &pD3DTexture)))
    {
        return rzErrFail;
    }
    else
    {
        // Put the alpha texture in the alpha channel of the first
        D3DLOCKED_RECT rcSrc;
        pD3DTexture->LockRect(0, &rcSrc, 0, 0);
        D3DLOCKED_RECT rcDst;
        m_pD3DTextureLogo->LockRect(0, &rcDst, 0, 0);
        RGBQUAD *pSrc = (RGBQUAD*)rcSrc.pBits, *pDst = (RGBQUAD*)rcDst.pBits;

        for (rzLong i = 0; i<m_dTextureLogoWidth*m_dTextureLogoHeight; i++)
        {
            pDst->rgbReserved = pSrc->rgbGreen;
            pDst++;
            pSrc++;
        }
        m_pD3DTextureLogo->UnlockRect(0);
        pD3DTexture->UnlockRect(0);
    }
    pD3DTexture->Release();

    // Renderer
    if (FAILED(m_pD3DDevice->CreateRenderTarget(m_dWidth, m_dHeight, D3DFMT_A8R8G8B8, D3DMULTISAMPLE_NONE, 0, FALSE, &m_pD3DScreen, 0)))
    {
        return rzErrFail;
    }

    BITMAPINFO bmi;
    bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bmi.bmiHeader.biWidth = m_dWidth;
    bmi.bmiHeader.biHeight = -m_dHeight;
    bmi.bmiHeader.biPlanes = 1;
    bmi.bmiHeader.biBitCount = 32;
    bmi.bmiHeader.biCompression = BI_RGB;
    bmi.bmiHeader.biSizeImage = 0;
    bmi.bmiHeader.biXPelsPerMeter = 0;
    bmi.bmiHeader.biYPelsPerMeter = 0;
    bmi.bmiHeader.biClrUsed = 0;
    bmi.bmiHeader.biClrImportant = 0;
    m_hBitmap = CreateDIBSection(NULL, &bmi, DIB_RGB_COLORS, &m_pBitmap, 0, 0);

    // Music
    FSOUND_File_SetCallbacks(memopen, memclose, memread, memseek, memtell);
#ifdef FMUSIC_ALL_ACTIVE
    m_pMod = FMUSIC_LoadSong("module.xm", NULL);
#else
    m_pMod = FMUSIC_LoadSong(0, NULL);
#endif
    return rzErrOk;
}

/*************************************************************************************************
 *
 * Process the demo
 */
rzBool rzrDemo::Process()
{
    if (m_bFirstFrame)
    {
        if (m_pMod)
            FMUSIC_PlaySong(m_pMod);

        QueryPerformanceCounter(&m_qPerfCounter);
        m_bFirstFrame = false;

        // Set the window on top of all and show it
        SetWindowPos(m_hWnd, HWND_TOPMOST, 0, 0, 0, 0,SWP_NOMOVE|SWP_NOSIZE|SWP_SHOWWINDOW);
    }
    else
    {
        rzQword qPerfCounter;
        QueryPerformanceCounter(&qPerfCounter);
        m_fTime = (rzFloat)((qPerfCounter.QuadPart - m_qPerfCounter.QuadPart)/(rzDouble)m_qPerfFrequency.QuadPart);
    }

    m_pD3DDevice->SetRenderTarget(0, m_pD3DScreen);
    m_pD3DDevice->SetRenderState(D3DRS_COLORWRITEENABLE, 0xf);

    m_pD3DDevice->BeginScene();
    m_pD3DDevice->SetVertexDeclaration(m_pD3DVD);

    m_pD3DDevice->SetTexture(0, m_pD3DTextureLogo);
    m_pD3DDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
    m_pD3DDevice->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);
    m_pD3DDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
    m_pD3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP);
    m_pD3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP);
    m_pD3DDevice->SetRenderState(D3DRS_ZENABLE, FALSE);
    m_pD3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

    m_pD3DDevice->SetPixelShader(m_pD3DPS);

    Effect();

    m_pD3DDevice->EndScene();

    //------------------------
    // Capture and set it transparent
    HDC dcScreen = ::GetDC(NULL);
    HDC dcMemory = ::CreateCompatibleDC(dcScreen);

    IDirect3DSurface9* pD3DScreen;
    m_pD3DDevice->CreateOffscreenPlainSurface(m_dWidth, m_dHeight, D3DFMT_A8R8G8B8, D3DPOOL_SYSTEMMEM, &pD3DScreen, 0);
    m_pD3DDevice->GetRenderTargetData(m_pD3DScreen, pD3DScreen);

    D3DLOCKED_RECT d3dRect;
    pD3DScreen->LockRect(&d3dRect, 0, 0);
    memcpy(m_pBitmap, d3dRect.pBits, m_dWidth*m_dHeight*4);
    pD3DScreen->UnlockRect();
    pD3DScreen->Release();

    SelectObject(dcMemory, m_hBitmap);

    RECT rcWindow;
    GetWindowRect(m_hWnd, &rcWindow);

    SIZE szDest = {m_dWidth, m_dHeight};
    POINT ptDest = {rcWindow.left, rcWindow.top};
    POINT ptSrc = {0, 0};

    BLENDFUNCTION blendPixelFunction= {AC_SRC_OVER, 0, 255, AC_SRC_ALPHA};
    UpdateLayeredWindow(m_hWnd, dcScreen, &ptDest, &szDest, dcMemory, &ptSrc, 0, &blendPixelFunction, ULW_ALPHA);

    ReleaseDC(m_hWnd, dcScreen);
    DeleteDC(dcMemory);

    return m_fTime<KeyFrame2?false:true;
}

/*************************************************************************************************
 *
 * Effect
 */
void rzrDemo::Effect()
{
#define stripes 512

    rzFloat fWidthRatio = m_dWidth/(rzFloat)m_dLogoWidth;
    rzFloat fHeightRatio = m_dHeight/(rzFloat)m_dLogoHeight;

    rzFloat fStartu, fStartv;
    rzFloat fWidth, fHeight;
    if (fWidthRatio<=fHeightRatio)
    {
        fWidth = 1.f;
        fStartu = 0.f;

        fHeight = m_dLogoHeight*fWidthRatio;
        fStartv = (fHeight*0.5f - m_dHeight*0.5f)/fHeight;
        fHeight = m_dHeight/fHeight;
    }
    else
    {
        fWidth = m_dLogoWidth*fHeightRatio;
        fStartu = (fWidth*0.5f - m_dWidth*0.5f)/fWidth;
        fWidth = m_dWidth*fWidthRatio;

        fHeight = 1.f;
        fStartv = 0.f;
    }

    rzFloat fRandScale;

    rzVertex vtx[4];
    if (m_fTime<KeyFrame0)
    {
        rzFloat fFade = m_fTime*m_fTime/(KeyFrame0*KeyFrame0);
        for (rzDword i = 0; i<4; i++)
        {
            vtx[i].pos.z = 0.5f;
            vtx[i].pos.w = 1.f;
            vtx[i].fade = fFade;
        }
        fRandScale = (KeyFrame0 + (KeyFrame1 - KeyFrame0)*0.2f - m_fTime)*0.15f;
    }
    else if (m_fTime<KeyFrame1)
    {
        for (rzDword i = 0; i<4; i++)
        {
            vtx[i].pos.z = 0.5f;
            vtx[i].pos.w = 1.f;
            vtx[i].fade = 1.f;
        }
        fRandScale = (KeyFrame0 + (KeyFrame1 - KeyFrame0)*0.2f - m_fTime)*0.15f;
        if (fRandScale<0)
            fRandScale = 0.f;
    }
    else if (m_fTime<KeyFrame2)
    {
        rzFloat fFade = (KeyFrame2 - m_fTime)*(KeyFrame2 - m_fTime)/((KeyFrame2 - KeyFrame1)*(KeyFrame2 - KeyFrame1));
        for (rzDword i = 0; i<4; i++)
        {
            vtx[i].pos.z = 0.5f;
            vtx[i].pos.w = 1.f;
            vtx[i].fade = fFade;
        }
        fRandScale = 0.f;
    }
    else
    {
        for (rzDword i = 0; i<4; i++)
        {
            vtx[i].pos.z = 0.5f;
            vtx[i].pos.w = 1.f;
            vtx[i].fade = 0;
        }
        fRandScale = 0.f;
    }

    for (rzDword y = 0; y<stripes; y++)
    {
        vtx[0].pos.x = -0.5f;
        vtx[0].pos.y = -0.5f + y*m_dHeight/stripes;

        vtx[1].pos.x = -0.5f + m_dWidth;
        vtx[1].pos.y = -0.5f + y*m_dHeight/stripes;

        vtx[2].pos.x = -0.5f;
        vtx[2].pos.y = -0.5f + (y + 1)*m_dHeight/stripes;

        vtx[3].pos.x = -0.5f + m_dWidth;
        vtx[3].pos.y = -0.5f + (y + 1)*m_dHeight/stripes;

        rzFloat fDisplace = fabsf(y - m_fTime*250);
        fDisplace = 35.f*(stripes*0.5f - fmodf(fDisplace, stripes))/stripes;
        rzFloat fInterference = 0.05f/(fDisplace*fDisplace + 1.f);

        vtx[0].coord[0].u = fStartu + (fRandScale + fInterference)*(0.5f - rand()/rzFloat(RAND_MAX));
        vtx[0].coord[1].u = fStartu + (fRandScale + fInterference)*(0.5f - rand()/rzFloat(RAND_MAX));
        vtx[0].coord[2].u = fStartu + (fRandScale + fInterference)*(0.5f - rand()/rzFloat(RAND_MAX));

        vtx[1].coord[0].u = fStartu + fWidth + (fRandScale + fInterference)*(0.5f - rand()/rzFloat(RAND_MAX));
        vtx[1].coord[1].u = fStartu + fWidth + (fRandScale + fInterference)*(0.5f - rand()/rzFloat(RAND_MAX));
        vtx[1].coord[2].u = fStartu + fWidth + (fRandScale + fInterference)*(0.5f - rand()/rzFloat(RAND_MAX));

        vtx[2].coord[0].u = vtx[0].coord[0].u;
        vtx[2].coord[1].u = vtx[0].coord[1].u;
        vtx[2].coord[2].u = vtx[0].coord[2].u;

        vtx[3].coord[0].u = vtx[1].coord[0].u;
        vtx[3].coord[1].u = vtx[1].coord[1].u;
        vtx[3].coord[2].u = vtx[1].coord[2].u;

        vtx[0].coord[0].v = fStartv + y*fHeight/stripes;
        vtx[0].coord[1].v = fStartv + y*fHeight/stripes + fInterference*0.37f;
        vtx[0].coord[2].v = fStartv + y*fHeight/stripes - fInterference*0.37f;

        vtx[1].coord[0].v = vtx[0].coord[0].v;
        vtx[1].coord[1].v = vtx[0].coord[1].v;
        vtx[1].coord[2].v = vtx[0].coord[2].v;

        vtx[2].coord[0].v = fStartv + (y + 1)*fHeight/stripes;
        vtx[2].coord[1].v = fStartv + (y + 1)*fHeight/stripes + fInterference*0.37f;
        vtx[2].coord[2].v = fStartv + (y + 1)*fHeight/stripes - fInterference*0.37f;

        vtx[3].coord[0].v = vtx[2].coord[0].v;
        vtx[3].coord[1].v = vtx[2].coord[1].v;
        vtx[3].coord[2].v = vtx[2].coord[2].v;

        m_pD3DDevice->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, vtx, sizeof(rzVertex));
    }
}
